/** \file
	\brief TCP client implementation
*/

#include "main.h"
#include "tcp_client.h"
#include "rtl8019.h"
#include "arp.h"
#include "ip.h"
#include "tcp.h"
#include "socket.h"
#include "clock.h"

uint8_t u8_ClientState = TC_IDLE;
uint32_t u32_usedIP;
extern unsigned long ul_PseudoRandom;
uint8_t u8_timeout = 0;	///< counter for timeouts

extern char szTelnetBuffer[35]; 	///< used as data source for tcp client
extern unsigned char packet[MTU_SIZE];

const uint16_t RemotePort = 80;
const unsigned char RemoteIP[] = {192,168,2,11};	///< remote IP address used by tcp client
PROGMEM char HTTP_POST[] = {
	"POST /my_script.php HTTP/1.0\r\n"\
    "Host: me.myhost.com\r\n"
    "Content-type: application/x-www-form-urlencoded\r\n"
    "Content-length: %VA3\r\n\r\n"
	"s=%VA4\r\n"
	"Connection: Close\r\n"
    "\r\n"	
	"%END"
};

/* php script:
$string=$_POST['s'];
$n = sscanf($string, "Temp: %fC  RH: %d%%  %dhPa", $temperature, $humidity, $pressure);
$result = mysql_query("insert into logi_temp (temp,czas) values('$temperature',NOW())",$db);
...
*/

int8_t lastConnectionTime = -10;

/** \brief Menages state machine for TCP client. Limited to 1 connection at time.
*/
void TcpClientConnectProcess(void)
{
	unsigned char MAC[6];
	//printf("CLIENT %d\r", u8_ClientState);
	if (u8_ClientState == TC_IDLE)
	{
		uint8_t diff = mm - lastConnectionTime;	
		if (diff < 0)
			diff += 60;
		if (diff < 5)	//limit connections to one per 5 minutes
			return;
		lastConnectionTime = mm;
	
		//start connection establishing
		//printf("Starting TCP connection...\r");
		uint32_t u32_myIP = IpCharToU32(MYIP);
		uint32_t u32_IP = IpCharToU32(RemoteIP);
		uint32_t mask = 0;
		for(uint8_t i=0; i<u8_netmask; i++)
			mask += (((uint32_t)1)<<i);
		
		if ((u32_myIP & mask) != (u32_IP & mask)) 
		{
			u32_usedIP = IpCharToU32(GateWay);
		}
		else
		{
			u32_usedIP = u32_IP;
		}
		
		if (ARP_Find(u32_usedIP, MAC) == 0)
		{
			//printf("Sending ARP req\n");
			SendArpRequest(u32_usedIP);
			u8_ClientState = TC_ARP_SENT;
			u8_timeout = 0;
		}
		else
		{
			u8_ClientState = TC_ARP_SENT;
			u8_timeout = 0;
		}
	}
	else if(u8_ClientState == TC_ARP_SENT)
	{
		if (++u8_timeout >= 5)
		{
			//printf("ARP reply timeout\r");
			u8_ClientState = TC_IDLE;
			return;
		}
		if (ARP_Find(u32_usedIP, MAC) == 0)
		{
			//printf("Resend ARP req\r");
			SendArpRequest(u32_usedIP);
			u8_ClientState = TC_ARP_SENT;
		}
		else
		{
		    //printf ("\nMac from ARP: %x.%x.%x.%x.%x.%x\n",MAC[0],MAC[1],MAC[2],
			//	MAC[3],MAC[4],MAC[5]);
			//prepare and send SYN		
		
		    struct TCP_Header *tcp;
			tcp = (struct TCP_Header *)&packet[TCP_OFFSET];
			tcp->TCP_SrcPort = SOCKET_CLIENT;
			tcp->TCP_DestPort = htons(RemotePort);
			memcpy(tcp->TCP_Seqnum, (char*)&ul_PseudoRandom, 4);
			tcp->TCP_HdrFlags[0] = 0x50;
			tcp->TCP_HdrFlags[1] = SYN_FLAG;
			tcp->TCP_Window = htons(MTU_SIZE);
			
			//make IP header
			unsigned int result16;
			struct Ethernet_Header *ethernet;
			ethernet = (struct Ethernet_Header *)&packet[ETHER_OFFSET];

			for (unsigned char a = 0; a < 6; a++)
			{
				//MAC Destadresse wird geschrieben mit MAC Sourceadresse
				ethernet->EnetPacketDest[a] = MAC[a];
				//Meine MAC Adresse wird in Sourceadresse geschrieben
				ethernet->EnetPacketSrc[a] = mymac[a];
			}
			ethernet->EnetPacketType = 0x0008;

			struct IP_Header *ip;
			ip = (struct IP_Header *)&packet[IP_OFFSET];
			
			ip->IP_Destaddr	= IpCharToU32(RemoteIP);
			ip->IP_Srcaddr	= *((unsigned long *)&MYIP[0]);
			ip->IP_Vers_Len = 0x45;
			ip->IP_Tos = 0x00;
			ip->IP_Pktlen = htons(40);		//20B IP header + 20B TCP header
			ip->IP_Frag_Offset = 0x40;		//do not fragment
			ip->IP_ttl = 128;
			ip->IP_Proto = PROT_TCP;
		
	
			//Berechnung der IP Checksumme
			//Alle Daten im IP Header werden addiert checksum wird deshalb
			//ersteinmal auf null gesetzt
			ip->IP_Hdr_Cksum = 0x0000;
			
			//Berechnung der IP Header lnge	
			result16 = (ip->IP_Vers_Len & 0x0F) << 2;

			//jetzt wird die Checksumme berechnet
			result16 = checksum (&ip->IP_Vers_Len, result16, 0);

			//schreibt Checksumme ins Packet
			ip->IP_Hdr_Cksum = ((result16 & 0xFF00) >> 8)|((result16 & 0x00FF)<<8);			
			
			//Berechnung der TCP Checksumme
			//Alle Daten im TCP Header werden addiert checksum wird deshalb
			//ersteinmal auf null gesetzt
			packet[TCP_CHKSUM] = 0x00;
			packet[TCP_CHKSUM+1] = 0x00;

			//Berechnet Headerlnge und Addiert Pseudoheaderlnge 2XIP = 8
			result16 = (packet[IP_PKTLEN]<<8) + packet[IP_PKTLEN+1] + 8;
			result16 = result16 - ((packet[IP_VERS_LEN] & 0x0F) << 2);
			unsigned long result32;
			result32 = result16 - 2;

			//Pointer wird auf das erste buffer im IP Header gesetzt
			//Routine berechnet die Checksumme
			result16 = checksum (&packet[IP_SRCADDR], result16, result32);

			//schreibt checksum ins buffer (INT H Byte)
			packet[TCP_CHKSUM] = ((result16 & 0xFF00) >> 8);
			//schreibt checksum ins buffer (INT L Byte)
			packet[TCP_CHKSUM+1] = (result16 & 0x00FF);
			
			//printf("Send SYN\n");
			Write_Ethernet_Frame (packet, ETHER_MIN_FRAME_SIZE);	//???
			u8_ClientState = TC_SYN_SENT;
		}	
	}
	else if(u8_ClientState == TC_DATA_SENT || u8_ClientState == TC_SYN_SENT)
	{
		if (++u8_timeout >= 5)
		{
			//printf("SYN or DATA ACK timeout\r");
			/** \todo RST? */
			u8_ClientState = TC_IDLE;
			return;
		}	
	}
}


void tcp_client (char* buffer,int *bufferlen)
{	
	if (u8_ClientState < TC_SYN_SENT)
	{
		//printf("rcv too early\n");	
		tcp_store(buffer);
		buffer[TCP_HDRFLAGS+1] = (RST_FLAG);
		tcp_make(buffer);			
		Write_Ethernet_Frame (buffer,*bufferlen);					
		return;	
	}
	if ((buffer[TCP_SRCPORT] == (RemotePort>>8)) &&
		(buffer[TCP_SRCPORT+1] == (RemotePort&0xFF)))
	{
		struct IP_Header *ip;
		ip = (struct IP_Header *)&buffer[IP_OFFSET];
		if (ip->IP_Srcaddr == IpCharToU32(RemoteIP))
		{
			if (u8_ClientState == TC_SYN_SENT)
			{
				//printf ("TC_SYN_SENT state\n");

				if (buffer[TCP_HDRFLAGS+1] == (ACK_FLAG|SYN_FLAG))
				{
					//printf ("ACK + SYN rcv\n");
					tcp_store(buffer);
					buffer[TCP_HDRFLAGS+1] = ACK_FLAG;
						
					//Erzeugt ein TCP buffer 
					tcp_make(buffer);			
					//Sendet das erzeugte TCP buffer 
					Write_Ethernet_Frame (buffer,*bufferlen);	

					//CONNECTION ESTABLISHED!
						
					buffer[TCP_HDRFLAGS+1] = ACK_FLAG | PSH_FLAG;
					tcp_add_data (HTTP_POST, buffer, bufferlen);
					tcp_make_client(buffer);
					Write_Ethernet_Frame (buffer,*bufferlen);
					
					buffer[TCP_HDRFLAGS+1] = ACK_FLAG | FIN_FLAG;
					tcp_add_data (HTTP_POST, buffer, bufferlen);
					tcp_make_client(buffer);
					Write_Ethernet_Frame (buffer,*bufferlen);

					u8_ClientState = TC_DATA_SENT;
					u8_timeout = 0;
				}
				else
				{
					printf("Connection not established\n");
					tcp_store(buffer);
					//buffer[TCP_HDRFLAGS+1] = (RST_FLAG);
					buffer[TCP_HDRFLAGS+1] = ACK_FLAG | FIN_FLAG | PSH_FLAG;
					tcp_make(buffer);			
					Write_Ethernet_Frame (buffer,*bufferlen);		
					u8_ClientState = TC_IDLE;
				}
			}
			else if (u8_ClientState == TC_DATA_SENT)
			{
				//we're expecting ACK
				//printf("TC_DATA_SENT state\n");
				if (buffer[TCP_HDRFLAGS+1] == (ACK_FLAG))
				{
					//printf("ACK after data sent\n");
					tcp_store(buffer);
					buffer[TCP_HDRFLAGS+1] = (ACK_FLAG | FIN_FLAG);
					tcp_make(buffer);			
					Write_Ethernet_Frame (buffer,*bufferlen);						
				}
				else
				{
					//printf("non ACK after data sent\n");
					tcp_store(buffer);
					buffer[TCP_HDRFLAGS+1] = (RST_FLAG);
					tcp_make(buffer);			
					Write_Ethernet_Frame (buffer,*bufferlen);										
				}
				u8_ClientState = TC_IDLE;
			}
			//else
			//	printf("Unexpected state %d\n", u8_ClientState);
		}
	}
}
